#include "MultiResolutionPathPlanner.h"
 
#include <cmath>
#include <queue>
#include <limits>
#include <cfloat>

MultiResolutionPathPlanner::MultiResolutionPathPlanner()
{
	grid = (GRID) new GRID_TYPE[GRID_R_DIM][GRID_Y_DIM][GRID_X_DIM];
	score = (GRID) new GRID_TYPE[GRID_R_DIM][GRID_Y_DIM][GRID_X_DIM];
	closed = (GRID) new GRID_TYPE[GRID_R_DIM][GRID_Y_DIM][GRID_X_DIM];
	for ( int i = 0; i < GRID_R_DIM; i++ )
		factor[i] = float( 1 << i );
	InitGrid();
	InitNeigbours();
}

void MultiResolutionPathPlanner::update( PathPlannerData& pathPlannerData )
{
	DECLARE_DEBUG_DRAWING( "module:MultiResolutionPathPlanner:path", "drawingOnField" );
	DECLARE_DEBUG_DRAWING( "module:MultiResolutionPathPlanner:obstacles", "drawingOnField" );
	DECLARE_DEBUG_DRAWING( "module:MultiResolutionPathPlanner:grid", "drawingOnField" );
	DECLARE_DEBUG_DRAWING( "module:MultiResolutionPathPlanner:expansions1", "drawingOnField" );
	DECLARE_DEBUG_DRAWING( "module:MultiResolutionPathPlanner:expansions2", "drawingOnField" );
	
	pathPlannerData.pathPlanned = false;
	
	if ( thePathPlannerControl.active == true ) {
		updateMultiResolutionPathPlanner( thePathPlannerControl.target.x, thePathPlannerControl.target.y, pathPlannerData );
		pathPlannerData.pathPlanned = true;
	}
}

void MultiResolutionPathPlanner::updateMultiResolutionPathPlanner( float goal_x, float goal_y, PathPlannerData& pathPlannerData ) {
	
	goal_x = goal_x / RESOLUTION;
	goal_y = goal_y / RESOLUTION;
	
	int target_cell_r, target_cell_x, target_cell_y;
	target_cell_r = target_cell_x = target_cell_y = 0;

	InitTarget( target_cell_r, target_cell_x, target_cell_y, goal_x, goal_y );
	InitGrid();
	InitObstacles();
	InitRobotObstacle();
	InsertObstacles();
	PlanPath( pathPlannerData, target_cell_r, target_cell_x, target_cell_y );
	DrawPath(pathPlannerData);
	DrawGrid();
}

void MultiResolutionPathPlanner::AddNeigbour( int r, int y, int x, int nr, int ny, int nx, float len_start, float len_end ) {
	edge* tmpEdge;
	int n;
	// check if edge is already in list
	for ( n = 0; n < num_neighbours[r][y][x]; n++ ) {	
		tmpEdge = &neighbours[r][y][x][n];
		if( tmpEdge->r == nr && tmpEdge->y == ny && tmpEdge->x == nx )
			return;
	}
	
	// if edge is not in list, put it in list
	tmpEdge = &neighbours[r][y][x][n];
	tmpEdge->r = nr;
	tmpEdge->y = ny;
	tmpEdge->x = nx;
	tmpEdge->len_start = len_start;
	tmpEdge->len_end = len_end;
	num_neighbours[r][y][x]++;
}

void MultiResolutionPathPlanner::AddEdge( int r, int y, int x, int nr, int ny, int nx, float len_start, float len_end ) {
	AddNeigbour( r, y, x, nr, ny, nx, len_start, len_end );
	AddNeigbour( nr, ny, nx, r, y, x, len_end, len_start );
}

void MultiResolutionPathPlanner::InitGrid() {
	bzero(grid, GRID_X_DIM * GRID_Y_DIM * GRID_R_DIM * sizeof(GRID_TYPE));
	bzero(score, GRID_X_DIM * GRID_Y_DIM * GRID_R_DIM * sizeof(GRID_TYPE));
	bzero(closed, GRID_X_DIM * GRID_Y_DIM * GRID_R_DIM * sizeof(GRID_TYPE));
}

void MultiResolutionPathPlanner::InitTarget( int& r, int& x, int& y, float target_x, float target_y ) {
	getGridPosition( r, x, y, target_x, target_y );
	(*grid)[r][y][x] = -1;
} 

void MultiResolutionPathPlanner::InitObstacles() {
	obstacles.clear();
	obstacle tmpObstacle;
	
	// goal posts
	for ( unsigned int i = 0; i < ( sizeof(theGoalPercept.posts) / sizeof(theGoalPercept.posts[0]) ); i++ ) {		
		if ( theGoalPercept.posts[i].perceptionType != GoalPost::NEVER_SEEN && theFrameInfo.getTimeSince(theGoalPercept.posts[i].timeWhenLastSeen) < 2000 ) {
			int r, x, y;
			getGridPosition( r, x, y, ( theGoalPercept.posts[i].positionOnField.x / RESOLUTION ), ( theGoalPercept.posts[i].positionOnField.y / RESOLUTION ) );
			tmpObstacle.r = r;
			tmpObstacle.x = x;
			tmpObstacle.y = y;
			tmpObstacle.cx = theGoalPercept.posts[i].positionOnField.x;
			tmpObstacle.cy = theGoalPercept.posts[i].positionOnField.y;
			tmpObstacle.behind = false;
			tmpObstacle.costs = 2000;
			tmpObstacle.usgoal = true;
			
			obstacles.push_back( tmpObstacle );
		}
	}
	
	// robot positions
	for ( RobotsModel::RCIt it = theRobotsModel.robots.begin(); it != theRobotsModel.robots.end(); ++it)
	{
		int r, x, y;
		getGridPosition( r, x, y, ( it->relPosOnField.x / RESOLUTION ), ( it->relPosOnField.y / RESOLUTION ) );
		tmpObstacle.r = r;
		tmpObstacle.x = x;
		tmpObstacle.y = y;
		tmpObstacle.cx = it->relPosOnField.x;
		tmpObstacle.cy = it->relPosOnField.y;
		tmpObstacle.behind = false;
		tmpObstacle.costs = 2000;
		tmpObstacle.usgoal = false;
		
		obstacles.push_back( tmpObstacle );
	}
	
	// ultrasonic sensors
	// underestimate the measured distances
	float leftwidth, rightwidth, leftdist, rightdist;
	int distance = 1000; // threshold
	
	// left sensor	
	if ( theObstacleModel.distanceToLeftObstacle < distance ) {
		leftdist = abs( theObstacleModel.distanceToLeftObstacle * sin( 60 * M_PI / 180 ) );
		leftwidth = ceil( theObstacleModel.distanceToLeftObstacle / 2 );
		
		// only take every 50th point on the segment
		for ( unsigned int i = 0; i <= ( 2 * leftwidth ); i+=50 ) {
			float x, y;
			
			x = cos( 25 * M_PI / 180 ) * leftdist - sin( 25 * M_PI / 180 ) * ( i - leftwidth ) + 50; // 50 and 40 are the positions of the sonars on the robot
			y = sin( 25 * M_PI / 180 ) * leftdist + cos( 25 * M_PI / 180 ) * ( i - leftwidth ) - 40;
			
			int obsr, obsx, obsy;
			getGridPosition( obsr, obsx, obsy, ( x / RESOLUTION ), ( y / RESOLUTION ) );
			tmpObstacle.r = obsr;
			tmpObstacle.x = obsx;
			tmpObstacle.y = obsy;
			tmpObstacle.cx = x;
			tmpObstacle.cy = y;
			tmpObstacle.behind = false;
			tmpObstacle.costs = 12000;
			tmpObstacle.usgoal = true;
			
			obstacles.push_back( tmpObstacle );
		}
	}
		
	// right sensor
	if ( theObstacleModel.distanceToRightObstacle < distance ) {
		rightdist = abs( theObstacleModel.distanceToRightObstacle * sin( 60 * M_PI / 180 ) );
		rightwidth = ceil( theObstacleModel.distanceToRightObstacle / 2 );		
		
		// only take every 50th point on the segment
		for ( unsigned int i = 0; i <= ( 2 * rightwidth ); i+=50 ) {
			float x, y;
			
			x = cos( -25 * M_PI / 180 ) * rightdist - sin( -25 * M_PI / 180 ) * ( i - rightwidth ) + 50; // 50 and 40 are the positions of the sonars on the robot
			y = sin( -25 * M_PI / 180 ) * rightdist + cos( -25 * M_PI / 180 ) * ( i - rightwidth ) + 40;
			
			int obsr, obsx, obsy;
			getGridPosition( obsr, obsx, obsy, ( x / RESOLUTION ), ( y / RESOLUTION ) );
			tmpObstacle.r = obsr;
			tmpObstacle.x = obsx;
			tmpObstacle.y = obsy;
			tmpObstacle.cx = x;
			tmpObstacle.cy = y;
			tmpObstacle.behind = false;
			tmpObstacle.costs = 12000;
			tmpObstacle.usgoal = true;
			
			obstacles.push_back( tmpObstacle );
		}
	}
}

void MultiResolutionPathPlanner::InitRobotObstacle() {
	int size = 2;
	obstacle tmpObstacle;
	
	// raise costs with regard to the robots movement direction and speed
	float raiseFactor, length, threshold;
	float speedX = theMotionInfo.walkRequest.speed.translation.x;
	float speedY = theMotionInfo.walkRequest.speed.translation.y;
	int costs = 100;
		
	if ( speedX != 0 || speedY != 0 ) {
		length = sqrt( speedX * speedX + speedY * speedY );
		threshold = 45; // degrees
		
		if ( ( speedX != 0 && speedY == 0 ) || ( speedX == 0 && speedY != 0 ))
			raiseFactor = abs( speedX ) + abs( speedY );
		else
			raiseFactor = length;
		
		for ( int i = -size; i <= size; i++ ) {
			for ( int j = -size; j <= size; j++ ) {
				if ( i != 0 || j != 0 ) {
					float scalar = ( -speedX * i ) + ( -speedY * j );
					float incos = scalar / ( length * sqrt( i * i + j * j ) );
					float alpha = acos( incos );
					alpha = ( alpha * 180 ) / M_PI;
					
					if ( alpha <= threshold ) {
						int r, x, y;
						int mid_i, mid_j;
						mid_i = i + int( 0.5f + GRID_X_CENTER );
						mid_j = j + int( 0.5f + GRID_Y_CENTER );
						
						getGridPosition( r, x, y, i, j );
						
						tmpObstacle.r = r;
						tmpObstacle.x = x;
						tmpObstacle.y = y;
						tmpObstacle.behind = true;
						tmpObstacle.costs = ( costs * raiseFactor );
						tmpObstacle.usgoal = false;
			
						obstacles.push_back( tmpObstacle );
					}
				}
			}
		}				
	}
	
}

void MultiResolutionPathPlanner::InsertObstacles() {
	obstacle tmpObstacle;
	
	// expand obstacles
	int size, samecostssize, completesize;
	int length = obstacles.size();
	float decrease;
	int decrfactor;
	const int robotsize = int( 0.5f + ROBOTRADIUS / RESOLUTION );
	
	for ( int i = 0; i < length; i++ ) {
		if ( obstacles[i].behind == false ) {
			
			// different radii
			if ( obstacles[i].usgoal == true ) {
				size = int( 0.5f + USGOALRADIUS / RESOLUTION );
				samecostssize = size + robotsize;
				completesize = samecostssize + robotsize;
			}
			else {
				size = int( 0.5f + ROBOTRADIUS / RESOLUTION );
				samecostssize = 2 * size;
				completesize = 3 * size;
			}
			
			for ( int j = -completesize; j <= completesize; j++ ) {
				for ( int k = -completesize; k <= completesize; k++ ) {
					float nx, ny;
					nx = ( obstacles[i].cx + j * RESOLUTION ) / RESOLUTION;
					ny = ( obstacles[i].cy + k * RESOLUTION ) / RESOLUTION;
					
					if ( nx < -( ( GRID_X_DIM / 2 ) * factor[GRID_R_DIM-1] ) || nx >= ( ( GRID_X_DIM / 2 ) * factor[GRID_R_DIM-1] ) || ny < -( ( GRID_Y_DIM / 2 ) * factor[GRID_R_DIM-1] ) || ny >= ( ( GRID_Y_DIM / 2 ) * factor[GRID_R_DIM-1] ) )
						continue;
					if ( j == 0 && k == 0 )
						continue;
					
					int r, x, y;
					getGridPosition( r, x, y, nx, ny );
					tmpObstacle.r = r;
					tmpObstacle.x = x;
					tmpObstacle.y = y;
					tmpObstacle.cx = obstacles[i].cx + j * RESOLUTION;
					tmpObstacle.cy = obstacles[i].cy + k * RESOLUTION;
					tmpObstacle.behind = true;
					
					if ( j < -samecostssize || j > samecostssize || k < -samecostssize || k > samecostssize ) {
						// linear decreasing costs
						if ( abs(j) >= abs(k))
							decrfactor = fabs(j) - samecostssize;
						else
							decrfactor = fabs(k) - samecostssize;
						decrease = obstacles[i].costs / robotsize;
						tmpObstacle.costs = obstacles[i].costs - decrfactor * decrease;
					} 
					else {
						// same costs
						tmpObstacle.costs = obstacles[i].costs;
					}
				
					obstacles.push_back( tmpObstacle );
				}
			}
		}
	}
	
	// insert obstacles into grid
	for ( unsigned int i = 0; i < obstacles.size(); i++ ) {
		if ( obstacles[i].r < 0 || obstacles[i].r >= GRID_R_DIM || obstacles[i].x < 0 || obstacles[i].x >= GRID_X_DIM || obstacles[i].y < 0 || obstacles[i].y >= GRID_Y_DIM )
			continue;
		
		(*grid)[obstacles[i].r][obstacles[i].y][obstacles[i].x] += obstacles[i].costs / ( factor[obstacles[i].r] * factor[obstacles[i].r] );
	}

	COMPLEX_DRAWING("module:MultiResolutionPathPlanner:obstacles", {
		for(size_t i = 0; i < obstacles.size(); i++)
		{
			if ( obstacles[i].r < 0 || obstacles[i].r >= GRID_R_DIM || obstacles[i].x < 0 || obstacles[i].x >= GRID_X_DIM || obstacles[i].y < 0 || obstacles[i].y >= GRID_Y_DIM )
				continue;
			if (obstacles[i].costs < 1.f)
				continue;
			int color = ( obstacles[i].costs / ( factor[obstacles[i].r] * factor[obstacles[i].r] ) * 3 ) / 10;

			if ( color > 235 )
				color = 235;

			color = 255-color;
		
			float x, y;
			getCartPosition( x, y, obstacles[i].r, obstacles[i].x, obstacles[i].y );

			LARGE_DOT("module:MultiResolutionPathPlanner:obstacles", x, y, ColorRGBA( color, color, color ), ColorRGBA( color, color, color ));
		}
	});
}

void MultiResolutionPathPlanner::InitNeigbours() {
	int r, y, x;
	for ( r = 0; r < GRID_R_DIM; r++ ) {
		for ( y = 0; y < GRID_Y_DIM; y++ )
			for ( x = 0; x < GRID_X_DIM; x++ ) {
				num_neighbours[r][y][x] = 0;
				
				// iterate over all neighbors in an 8-neighborhood
				for ( int dy = -1; dy <= 1; dy++ )
					for ( int dx = -1; dx <= 1; dx++ ) {
						
						if ( dx == 0 && dy == 0 ) // exclude yourself
							continue;
						int ny = y + dy;
						int nx = x + dx;

 						if (
							( r == 0 && nx >= 0 && nx < GRID_X_DIM && ny >= 0 && ny < GRID_Y_DIM )
							|| 
							( ( nx >= 0 && nx < GRID_X_DIM && ny >= 0 && ny < GRID_Y_DIM )
							&& !( nx > GRID_X_MAX && nx < GRID_X_MIN && ny > GRID_Y_MAX && ny < GRID_Y_MIN )
							&& (x >= 0 && x < GRID_X_DIM && y >= 0 && y < GRID_Y_DIM )
							&& !( x > GRID_X_MAX && x < GRID_X_MIN && y > GRID_Y_MAX && y < GRID_Y_MIN )
							) 
						   ) {
							AddNeigbour( r, y, x, 
											r,
											ny,
											nx,
											factor[r] * float( sqrt( dx * dx + dy * dy ) ) / 2,
											factor[r] * float( sqrt( dx * dx + dy * dy ) ) / 2 );
											
						} 
						
					}
			}
			if ( r > 0 ) {
				for ( int x = GRID_X_MAX + 1; x < GRID_X_MIN; x++ ) {
					AddEdge( r, GRID_Y_MAX, x, 
											r - 1,
											0,	
											( x - GRID_X_QUARTER ) * 2,
											factor[r-1] * 2 * float( sqrt( 2.5f ) ) / 3,
											factor[r-1] * float( sqrt( 2.5f ) ) / 3 );
					AddEdge( r, GRID_Y_MAX, x, 
											r - 1,
											0,	
											( x - GRID_X_QUARTER ) * 2 + 1,
											factor[r-1] * 2 * float( sqrt( 2.5f ) ) / 3,
											factor[r-1] * float( sqrt( 2.5f ) ) / 3 );
					AddEdge( r, GRID_Y_MIN, x, 
											r - 1,
											GRID_Y_DIM - 1,	
											( x - GRID_X_QUARTER ) * 2,
											factor[r-1] * 2 * float( sqrt( 2.5f ) ) / 3,
											factor[r-1] * float( sqrt( 2.5f ) ) / 3 );
					AddEdge( r, GRID_Y_MIN, x, 
											r - 1,
											GRID_Y_DIM - 1,	
											( x - GRID_X_QUARTER ) * 2 + 1,
											factor[r-1] * 2 * float( sqrt( 2.5f ) ) / 3,
											factor[r-1] * float( sqrt( 2.5f ) ) / 3 );
				}
				
				for ( int y = GRID_Y_MAX + 1; y < GRID_Y_MIN; y++ ) {
					AddEdge( r, y, GRID_X_MAX, 
											r - 1,
											( y - GRID_Y_QUARTER ) * 2,
											0,	
											factor[r-1] * 2 * float( sqrt( 2.5f ) ) / 3,
											factor[r-1] * float( sqrt( 2.5f ) ) / 3 );
					AddEdge( r, y, GRID_X_MAX, 
											r - 1,
											( y - GRID_Y_QUARTER ) * 2 + 1,
											0,	
											factor[r-1] * 2 * float( sqrt( 2.5f ) ) / 3,
											factor[r-1] * float( sqrt( 2.5f ) ) / 3 );
					AddEdge( r, y, GRID_X_MIN, 
											r - 1,
											( y - GRID_Y_QUARTER ) * 2,
											GRID_X_DIM - 1,	
											factor[r-1] * 2 * float( sqrt( 2.5f ) ) / 3,
											factor[r-1] * float( sqrt( 2.5f ) ) / 3 );
					AddEdge( r, y, GRID_X_MIN, 
											r - 1,
											( y - GRID_Y_QUARTER ) * 2 + 1,
											GRID_X_DIM - 1,	
											factor[r-1] * 2 * float( sqrt( 2.5f ) ) / 3,
											factor[r-1] * float( sqrt( 2.5f ) ) / 3 );
				}
				
				AddEdge( r, GRID_Y_MAX, GRID_X_MAX, 
									r - 1,
									0,
									0,	
									factor[r-1] * 2 * float( sqrt( 4.5f ) ) / 3,
									factor[r-1] * float( sqrt( 4.5f ) ) / 3 );
				AddEdge( r, GRID_Y_MIN, GRID_X_MAX, 
									r - 1,
									GRID_Y_DIM - 1,
									0,	
									factor[r-1] * 2 * float( sqrt( 4.5f ) ) / 3,
									factor[r-1] * float( sqrt( 4.5f ) ) / 3 );
				AddEdge( r, GRID_Y_MAX, GRID_X_MIN, 
									r - 1,
									0,	
									GRID_X_DIM - 1,
									factor[r-1] * 2 * float( sqrt( 4.5f ) ) / 3,
									factor[r-1] * float( sqrt( 4.5f ) ) / 3 );
				AddEdge( r, GRID_Y_MIN, GRID_X_MIN, 
									r - 1,
									GRID_Y_DIM - 1,
									GRID_X_DIM - 1,
									factor[r-1] * 2 * float( sqrt( 4.5f ) ) / 3,
									factor[r-1] * float( sqrt( 4.5f ) ) / 3 );
		
				for ( int x = GRID_X_MAX + 1; x < GRID_X_MIN; x++ ) {
					if ( x <GRID_X_MIN - 1 ) {	
						AddEdge( r, GRID_Y_MAX, x, 
											r - 1,
											0,	
											( x - GRID_X_QUARTER ) * 2 + 2,
											factor[r-1] * 2 * float( sqrt( 4.5f ) ) / 3,
											factor[r-1] * float( sqrt( 4.5f ) ) / 3 );

						AddEdge( r, GRID_Y_MIN, x, 
											r - 1,
											GRID_Y_DIM-1,	
											( x - GRID_X_QUARTER ) * 2 + 2,
											factor[r-1] * 2 * float( sqrt( 4.5f ) ) / 3,
											factor[r-1] * float( sqrt( 4.5f ) ) / 3 );
					}
					if ( x > GRID_X_MAX + 1 ) {
						AddEdge( r, GRID_Y_MAX, x, 
											r - 1,
											0,	
											( x - GRID_X_QUARTER ) * 2 - 1,
											factor[r-1] * 2 * float( sqrt( 4.5f ) ) / 3,
											factor[r-1] * float( sqrt( 4.5f ) ) / 3 );
						AddEdge( r, GRID_Y_MIN, x, 
											r - 1,
											GRID_Y_DIM - 1,	
											( x - GRID_X_QUARTER ) * 2 - 1,
											factor[r-1] * 2 * float( sqrt( 4.5f ) ) / 3,
											factor[r-1] * float( sqrt( 4.5f ) ) / 3 );
					}
				}
				
				for ( int y = GRID_Y_MAX + 1; y < GRID_Y_MIN; y++ ) {
					if ( y > GRID_Y_MAX + 1 ) {
						AddEdge( r, y, GRID_X_MAX, 
											r - 1,
											( y - GRID_Y_QUARTER ) * 2 - 1,
											0,	
											factor[r-1] * 2 * float( sqrt( 4.5f ) ) / 3,
											factor[r-1] * float( sqrt( 4.5f ) ) / 3 );
						AddEdge( r, y, GRID_X_MIN, 
											r - 1,
											( y - GRID_Y_QUARTER ) * 2 - 1,
											GRID_X_DIM - 1,	
											factor[r-1] * 2 * float( sqrt( 4.5f ) ) / 3,
											factor[r-1] * float( sqrt( 4.5f ) ) / 3 );
					}
					if ( y < GRID_Y_MIN - 1 ) {
						AddEdge( r, y, GRID_X_MAX, 
											r - 1,
											( y - GRID_Y_QUARTER ) * 2 + 2,
											0,	
											factor[r-1] * 2 * float( sqrt( 4.5f ) ) / 3,
											factor[r-1] * float( sqrt( 4.5f ) ) / 3 );
						AddEdge( r, y, GRID_X_MIN, 
											r - 1,
											( y - GRID_Y_QUARTER ) * 2 + 2,
											GRID_X_DIM - 1,	
											factor[r-1] * 2 * float( sqrt( 4.5f ) ) / 3,
											factor[r-1] * float( sqrt( 4.5f ) ) / 3 );
					}
				}
			}

	}
}

void MultiResolutionPathPlanner::PlanPath( PathPlannerData& pathPlannerData, int target_cell_r, int target_cell_x, int target_cell_y ) {	
	std::priority_queue< gridpoint > priorityqueue;
	std::vector< gridpoint > visited;
	
	gridpoint currPos;
	gridpoint tmpPos;
	gridpoint pathPos;
	
	float costs;
	float heuristic;
	float scores;
	bool reachedGoal = false;
	bool reachedStart = false;
	
	float target_x, target_y;
	getCartPosition( target_x, target_y, target_cell_r, target_cell_x, target_cell_y );
	
	// put first four grid cells manually in priority queue because of the robot standing in the middle between grid cells
	static const float costsInitialStep = sqrtf(2.) * RESOLUTION / 2.f;
	tmpPos.r = 0;
	tmpPos.x = int( GRID_X_CENTER + 0.5f );
	tmpPos.y = int( GRID_Y_CENTER + 0.5f );
	tmpPos.costs = costsInitialStep;
	tmpPos.scores = tmpPos.costs + sqrt( ( ( tmpPos.x - GRID_X_CENTER ) * RESOLUTION * factor[tmpPos.r] - target_x ) * ( ( tmpPos.x - GRID_X_CENTER ) * RESOLUTION * factor[tmpPos.r] - target_x ) + ( ( tmpPos.y - GRID_Y_CENTER ) * RESOLUTION * factor[tmpPos.r] - target_y ) * ( ( tmpPos.y - GRID_Y_CENTER ) * RESOLUTION * factor[tmpPos.r] - target_y ) );
	(*score)[tmpPos.r][tmpPos.y][tmpPos.x] = tmpPos.scores;
	priorityqueue.push( tmpPos );
	
	tmpPos.r = 0;
	tmpPos.x = int( GRID_X_CENTER - 0.5f );
	tmpPos.y = int( GRID_Y_CENTER + 0.5f );
	tmpPos.costs = costsInitialStep;
	tmpPos.scores = tmpPos.costs + sqrt( ( ( tmpPos.x - GRID_X_CENTER ) * RESOLUTION * factor[tmpPos.r] - target_x ) * ( ( tmpPos.x - GRID_X_CENTER ) * RESOLUTION * factor[tmpPos.r] - target_x ) + ( ( tmpPos.y - GRID_Y_CENTER ) * RESOLUTION * factor[tmpPos.r] - target_y ) * ( ( tmpPos.y - GRID_Y_CENTER ) * RESOLUTION * factor[tmpPos.r] - target_y ) );
	(*score)[tmpPos.r][tmpPos.y][tmpPos.x] = tmpPos.scores;
	priorityqueue.push( tmpPos );
	
	tmpPos.r = 0;
	tmpPos.x = int( GRID_X_CENTER + 0.5f );
	tmpPos.y = int( GRID_Y_CENTER - 0.5f );
	tmpPos.costs = costsInitialStep;
	tmpPos.scores = tmpPos.costs + sqrt( ( ( tmpPos.x - GRID_X_CENTER ) * RESOLUTION * factor[tmpPos.r] - target_x ) * ( ( tmpPos.x - GRID_X_CENTER ) * RESOLUTION * factor[tmpPos.r] - target_x ) + ( ( tmpPos.y - GRID_Y_CENTER ) * RESOLUTION * factor[tmpPos.r] - target_y ) * ( ( tmpPos.y - GRID_Y_CENTER ) * RESOLUTION * factor[tmpPos.r] - target_y ) );
	(*score)[tmpPos.r][tmpPos.y][tmpPos.x] = tmpPos.scores;
	priorityqueue.push( tmpPos );
	
	tmpPos.r = 0;
	tmpPos.x = int( GRID_X_CENTER - 0.5f );
	tmpPos.y = int( GRID_Y_CENTER - 0.5f );
	tmpPos.costs = costsInitialStep;
	tmpPos.scores = tmpPos.costs + sqrt( ( ( tmpPos.x - GRID_X_CENTER ) * RESOLUTION * factor[tmpPos.r] - target_x ) * ( ( tmpPos.x - GRID_X_CENTER ) * RESOLUTION * factor[tmpPos.r] - target_x ) + ( ( tmpPos.y - GRID_Y_CENTER ) * RESOLUTION * factor[tmpPos.r] - target_y ) * ( ( tmpPos.y - GRID_Y_CENTER ) * RESOLUTION * factor[tmpPos.r] - target_y ) );
	priorityqueue.push( tmpPos );
	
	while ( !priorityqueue.empty() ) {
		currPos = priorityqueue.top();
		priorityqueue.pop();
		visited.push_back( currPos );
		
		if ( currPos.r == target_cell_r && currPos.x == target_cell_x && currPos.y == target_cell_y ) {
			reachedGoal = true;
			break;
		}
		
		// expand node
		for ( int n = 0; n < num_neighbours[currPos.r][currPos.y][currPos.x]; n++) {
			const edge *np = &neighbours[currPos.r][currPos.y][currPos.x][n];
			
			if ( (*closed)[np->r][np->y][np->x] == 1 )
					continue;
			
			tmpPos.r = np->r;
			tmpPos.y = np->y;
			tmpPos.x = np->x;
			
			// calculate costs
			costs = currPos.costs + ( np->len_start*RESOLUTION + np->len_end*RESOLUTION ) + (*grid)[tmpPos.r][tmpPos.y][tmpPos.x];
			tmpPos.costs = costs;
			
			// calculate heuristic
			heuristic = sqrt( ( ( tmpPos.x - GRID_X_CENTER ) * RESOLUTION * factor[tmpPos.r] - target_x ) * ( ( tmpPos.x - GRID_X_CENTER ) * RESOLUTION * factor[tmpPos.r] - target_x ) + ( ( tmpPos.y - GRID_Y_CENTER ) * RESOLUTION * factor[tmpPos.r] - target_y ) * ( ( tmpPos.y - GRID_Y_CENTER ) * RESOLUTION * factor[tmpPos.r] - target_y ) );
			scores = costs + heuristic;
			
			// visualize expansions of A*
			// LINE("module:MultiResolutionPathPlanner:expansions1", ( currPos.x - GRID_X_CENTER ) * RESOLUTION * factor[currPos.r], ( currPos.y - GRID_Y_CENTER ) * RESOLUTION * factor[currPos.r], ( tmpPos.x - GRID_X_CENTER ) * RESOLUTION * factor[tmpPos.r], ( tmpPos.y - GRID_Y_CENTER ) * RESOLUTION * factor[tmpPos.r], 10, Drawings::ps_solid, ColorClasses::red);
			
			
			if ( ( (*score)[tmpPos.r][tmpPos.y][tmpPos.x] != 0 ) && ( (*score)[tmpPos.r][tmpPos.y][tmpPos.x] <= scores ) )
				continue;
			
			// visualize expansions of a* after testing scores
			// LINE("module:MultiResolutionPathPlanner:expansions2", ( currPos.x - GRID_X_CENTER ) * RESOLUTION * factor[currPos.r], ( currPos.y - GRID_Y_CENTER ) * RESOLUTION * factor[currPos.r], ( tmpPos.x - GRID_X_CENTER ) * RESOLUTION * factor[tmpPos.r], ( tmpPos.y - GRID_Y_CENTER ) * RESOLUTION * factor[tmpPos.r], 10, Drawings::ps_solid, ColorClasses::red);
			
			// parent node
			tmpPos.pr = currPos.r;
			tmpPos.px = currPos.x;
			tmpPos.py = currPos.y;
			
			(*score)[tmpPos.r][tmpPos.y][tmpPos.x] = scores;
			tmpPos.scores = scores;
			priorityqueue.push( tmpPos );
		}		
		(*closed)[currPos.r][currPos.y][currPos.x] = 1;
	}
	
	// backtrack path
	pathPlannerData.path.clear();
	if ( reachedGoal && !visited.empty()) {	
		pathPos = visited.back();
		visited.pop_back();

		float waypoint_x, waypoint_y;
		getCartPosition( waypoint_x, waypoint_y, pathPos.r, pathPos.x, pathPos.y );
		pathPlannerData.path.push_back( Vector2<float>(waypoint_x, waypoint_y) );
		
		while ( !reachedStart ) {
			float scores = FLT_MAX;
			for ( unsigned int i = 0; i < visited.size(); i++ ) {		
				if ( visited[i].r == pathPos.pr && visited[i].x == pathPos.px && visited[i].y == pathPos.py ) {
					if ( visited[i].scores < scores ) {
						scores = visited[i].scores;
						pathPos = visited[i];
					}
				}
			}
			
			getCartPosition( waypoint_x, waypoint_y, pathPos.r, pathPos.x, pathPos.y );
			pathPlannerData.path.push_back( Vector2<float>(waypoint_x, waypoint_y) );

			if ( pathPos.r == 0 && ( ( pathPos.x == int( 0.5f + GRID_X_CENTER ) && pathPos.y == int( 0.5f + GRID_Y_CENTER ) ) || ( pathPos.x == int( 0.5f + GRID_X_CENTER ) && pathPos.y == int( GRID_Y_CENTER ) ) || ( pathPos.x == int( GRID_X_CENTER ) && pathPos.y == int( 0.5f + GRID_Y_CENTER ) ) || ( pathPos.x == int( GRID_X_CENTER ) && pathPos.y == int( GRID_Y_CENTER ) ) ) ) {
				reachedStart = true;
				break;
			}
		}
	}
}
 
MAKE_MODULE(MultiResolutionPathPlanner, Modeling)
