/*!
 * \class MultiResolutionPathPlanner
 * \brief a resource efficient path planner using a local multiresolution grid
 * \author Ricarda Steffens (steffens _at_ cs.uni-bonn.de)
 * \author Matthias Nieuwenhuisen (nieuwenh _at_ ais.uni-bonn.de)
 * \date December 17, 2010
 */

#ifndef __MultiResolutionPathPlanner_h__
#define __MultiResolutionPathPlanner_h__

#include "Tools/Module/Module.h"

#include "Representations/Modeling/ObstacleModel.h"
#include "Representations/MotionControl/OdometryData.h"
#include "Representations/Perception/GoalPercept.h"
#include "Representations/Infrastructure/FrameInfo.h"
#include "Representations/MotionControl/MotionInfo.h"
#include "Representations/Modeling/RobotsModel.h"
#include "Representations/Modeling/PathPlannerData.h"
#include "Representations/Modeling/PathPlannerControl.h"

#include <cmath>

#define _USE_MATH_DEFINES

#define GRID_X_DIM  8	     // x-dimension of each of the grids, should be a multiple of 4
#define GRID_Y_DIM  8	     // y-dimension of each of the grids, should be a multiple of 4
#define GRID_X_CENTER  3.5f	 // GRID_X_DIM / 2 - 0.5
#define GRID_Y_CENTER  3.5f	 // GRID_Y_DIM / 2 - 0.5
#define GRID_X_QUARTER  2	 // GRID_X_DIM / 4
#define GRID_Y_QUARTER  2	 // GRID_Y_DIM / 4
#define GRID_X_MIN  6	     // interval  in which the grids with the higher resolution lie in
#define GRID_X_MAX  1
#define GRID_Y_MIN  6
#define GRID_Y_MAX  1
#define GRID_R_DIM  5	     // number of resolution levels (k=log(n/m)+1), m=8, n=128 (m is dimension of multiresolution grid, n is dimension of uniform grid covering the same area)

#define MAX_NEIGHBORS 9	// max number of neighbors

#define RESOLUTION 100.f	// smallest cellsize in mm
#define ROBOTRADIUS 150.f	// radius of the planning nao robot
#define USGOALRADIUS 55.f	// radius of the goal posts, also used for obstacles perceived by the ultrasonic sensors

MODULE(MultiResolutionPathPlanner)
	REQUIRES(ObstacleModel)
	REQUIRES(OdometryData)
	REQUIRES(GoalPercept)
	REQUIRES(FrameInfo)
	REQUIRES(MotionInfo)
	REQUIRES(RobotsModel)
	USES(PathPlannerControl)
	PROVIDES_WITH_MODIFY(PathPlannerData)
END_MODULE

class MultiResolutionPathPlanner : public MultiResolutionPathPlannerBase {
private:
	void update( PathPlannerData& pathPlannerData );

public:
	MultiResolutionPathPlanner();
	void updateMultiResolutionPathPlanner( float target_x, float target_y, PathPlannerData& pathPlannerData ); // target coordinates in mm

	typedef float GRID_TYPE;
	typedef GRID_TYPE (*GRID)[GRID_R_DIM][GRID_Y_DIM][GRID_X_DIM];
	GRID grid; // obstacles
	GRID score;	// costs + heuristics
	GRID closed; // visited cells
	
	float factor[GRID_R_DIM];

	void InitNeigbours();
	/** initialize neighbors and the corresponding edges */
	void AddNeigbour( int r, int y, int x, int nr, int ny, int nx, float len_start, float len_end );
	/** initialize edge between two neighboring cells from both sides */
	void AddEdge( int r, int y, int x, int nr, int ny, int nx, float len_start, float len_end );
	/** initialize each grid with zeros */
	void InitGrid();
	/** determine the target position in local multiresolution coordinates and initialize it in the grid */
	void InitTarget( int& target_cell_r, int& target_cell_x, int& target_cell_y, float target_x, float target_y );
	/** plan path to target with A* algorithm, backtracking */
	void PlanPath( PathPlannerData& pathPlannerData, int target_cell_r, int target_cell_x, int target_cell_y );
	/** Draws the multiresolution grid structure into the world state */
	void DrawGrid();
	/** Draws the found path into the world state */
	void DrawPath( const PathPlannerData& pathPlannerData );
	/** push the perceived goal posts, the positions of the robots and the measurements of both ultrasonic sensors into the obstacle vector */
	void InitObstacles();
	/** initialize the artificial motion obstacle behind the robot */
	void InitRobotObstacle();
	/** expand obstacles, inserts them into the grid and draw them into the world state */
	void InsertObstacles();
	
	/** conversion from multiresolution coordinates to cartesian coordinates */
	inline void getCartPosition ( float& pos_x, float& pos_y, int r, int x, int y ) {
		pos_x = ( x - GRID_X_CENTER ) * RESOLUTION * factor[r];
		pos_y = ( y - GRID_Y_CENTER ) * RESOLUTION * factor[r];
	}
	
	/** conversion from cartesian coordinates to multiresolution coordinates */
	void getGridPosition ( int& r, int& x, int& y, float pos_x, float pos_y );
	
	struct gridpoint {
		int r, x, y;
		int pr, px, py; // parent
		float costs;
		float scores;
		
		// operator for priority queue order
		bool operator<( const gridpoint &a ) const {
			return scores > a.scores;
		}
	};
	
	struct edge {
		int r, x, y;
		float len_start; // length from starting point to cellborder
		float len_end; // length from target point to cellborder
	};
	
	edge neighbours[GRID_R_DIM][GRID_Y_DIM][GRID_X_DIM][MAX_NEIGHBORS]; // adjacency list
	short num_neighbours[GRID_R_DIM][GRID_Y_DIM][GRID_X_DIM]; // number of neighbors for every gridcell
	
	struct obstacle {
		int r, x, y;
		int cx, cy; // cartesian coordinates
		float costs;
		bool behind; // is artificial obstacle behind the robot
		bool usgoal; // distinguish between goal posts/ultrasonic sensor measurements and robots
	};
	
	std::vector< obstacle > obstacles;
};

inline void MultiResolutionPathPlanner::getGridPosition ( int& r, int& x, int& y, float pos_x, float pos_y ) {
	r = GRID_R_DIM;

	do {
		r--;
		float factor = 1 / float( 1 << ( r ) );
		x = int( 0.5f + GRID_X_CENTER + pos_x * factor );
		y = int( 0.5f + GRID_Y_CENTER + pos_y * factor );
	}
	while ( r > 0 && 
			x < GRID_X_MIN && x > GRID_X_MAX && 
			y < GRID_Y_MIN && y > GRID_Y_MAX ); 
}

inline void MultiResolutionPathPlanner::DrawGrid() {
	COMPLEX_DRAWING("module:MultiResolutionPathPlanner:grid", {
		int fromX, fromY, toX, toY;
		float pos_x, pos_y;
		fromX = fromY = toX = toY = pos_x = pos_y = 0;
		float length;

		for ( int r = 0; r < GRID_R_DIM; r++ ) {
			for ( int y = 0; y < GRID_Y_DIM; y++ ) {
				for ( int x = 0; x < GRID_X_DIM; x++ ) {
					
					length = ( RESOLUTION * factor[r] ) / 2;
					getCartPosition( pos_x, pos_y, r, x, y );
					
					fromX = pos_x - length;
					fromY = pos_y - length;
					toX = pos_x - length;
					toY = pos_y + length;
					
					LINE("module:MultiResolutionPathPlanner:grid", fromX, fromY, toX, toY, 10, Drawings::ps_solid, ColorClasses::black);
					
					toX = pos_x + length;
					toY = pos_y - length;
					
					LINE("module:MultiResolutionPathPlanner:grid", fromX, fromY, toX, toY, 10, Drawings::ps_solid, ColorClasses::black);
					
					fromX = pos_x + length;
					fromY = pos_y + length;
					
					LINE("module:MultiResolutionPathPlanner:grid", fromX, fromY, toX, toY, 10, Drawings::ps_solid, ColorClasses::black);
					
					toX = pos_x - length;
					toY = pos_y + length;
					
					LINE("module:MultiResolutionPathPlanner:grid", fromX, fromY, toX, toY, 10, Drawings::ps_solid, ColorClasses::black);
				}
			}
		} 
		);
	}
}

inline void MultiResolutionPathPlanner::DrawPath( const PathPlannerData& pathPlannerData ) {
	COMPLEX_DRAWING("module:MultiResolutionPathPlanner:path", {
		if(pathPlannerData.path.empty())
			return;
		float lastX = pathPlannerData.path.front().x;
		float lastY = pathPlannerData.path.front().y;
		for( std::vector< Vector2<float> >::const_iterator it = pathPlannerData.path.begin(); it != pathPlannerData.path.end(); ++it )
		{
			LARGE_DOT("module:MultiResolutionPathPlanner:path", it->x, it->y, ColorClasses::red, ColorClasses::red);
			LINE("module:MultiResolutionPathPlanner:path", lastX, lastY, it->x, it->y, 10, Drawings::ps_solid, ColorClasses::black);
			lastX = it->x;
			lastY = it->y;
		}
	}
	);
};

#endif
