#include "PolargridPathPlanner.h"
 
#include <math.h>
#include <queue>
#include <limits>

PolargridPathPlanner::PolargridPathPlanner()
{
	// initialize grids
	grid = (GRID) new GRID_TYPE[GRID_RHO][GRID_THETA];
	score = (GRID) new GRID_TYPE[GRID_RHO][GRID_THETA];
	closed = (GRID) new GRID_TYPE[GRID_RHO][GRID_THETA];
	InitGrid();
	InitLookup();
}

void PolargridPathPlanner::update( PathPlannerData& pathPlannerData )
{
	DECLARE_DEBUG_DRAWING( "module:PolargridPathPlanner:path", "drawingOnField" );
	DECLARE_DEBUG_DRAWING( "module:PolargridPathPlanner:grid", "drawingOnField" );
	DECLARE_DEBUG_DRAWING( "module:PolargridPathPlanner:obstacles", "drawingOnField" );
	DECLARE_DEBUG_DRAWING( "module:PolargridPathPlanner:expansion", "drawingOnField" );
	
	pathPlannerData.pathPlanned = false;
	
	if( thePathPlannerControl.active == true ) {
		updatePolargridPathPlanner( thePathPlannerControl.target.x, thePathPlannerControl.target.y, pathPlannerData );
		pathPlannerData.pathPlanned = true;
	}
}

void PolargridPathPlanner::updatePolargridPathPlanner( float target_x, float target_y, PathPlannerData& pathPlannerData )
{
	int r = 0;
	int t = 0;
	
	InitGrid();
	InitTarget( r, t, target_x, target_y );
	InitObstacles();
	InitRobotObstacle();
	InsertObstacles();
	PlanPath( pathPlannerData, r, t );
	DrawPath( pathPlannerData );
	DrawGrid();
	
}

void PolargridPathPlanner::InitGrid() {
	bzero(grid, GRID_RHO * GRID_THETA * sizeof(GRID_TYPE));
	bzero(score, GRID_RHO * GRID_THETA * sizeof(GRID_TYPE));
	bzero(closed, GRID_RHO * GRID_THETA * sizeof(GRID_TYPE));
}

void PolargridPathPlanner::InitLookup() {
	for ( int r = 0; r <= GRID_RHO; r++ ) {
		lookuprho[r] = ( pow( BASE, r ) - 1 ) * OFFSET ;
		lookuprhohalf[r] = ( pow( BASE, r+0.5 ) - 1 ) * OFFSET;
	}
	for ( int t = 0; t < GRID_THETA; t++ ) {
		lookuptheta[t] = getTheta(t);
		lookupsintheta[t] = sin(getTheta(t));
		lookupcostheta[t] = cos(getTheta(t));
	}
	for(int r = 0; r <= GRID_RHO; r++)
	{
		for(int rdiff = 0; rdiff <= 1; rdiff++)
			for(int tdiff = 0; tdiff <= 1; tdiff++)
				if(rdiff != 0 || tdiff != 0)
					lookupcosts[r][rdiff+2*tdiff-1] = getCosts(r, 0, r+rdiff, tdiff);
	}
}

void PolargridPathPlanner::InitTarget( int& r, int& t, float target_x, float target_y ) {
	getPolarPosition( r, t, target_x, target_y );
	(*grid)[r][t] = -1;
} 

void PolargridPathPlanner::InitObstacles() {
	obstacles.clear();
	obstacle tmpObstacle;
	
	// goal posts
	for ( unsigned int i = 0; i < ( sizeof(theGoalPercept.posts) / sizeof(theGoalPercept.posts[0]) ); i++ )
	{
		if ( theGoalPercept.posts[i].perceptionType != GoalPost::NEVER_SEEN && theFrameInfo.getTimeSince(theGoalPercept.posts[i].timeWhenLastSeen) < 2000 )
		{
			int r, t;
			getPolarPosition( r, t, theGoalPercept.posts[i].positionOnField.x, theGoalPercept.posts[i].positionOnField.y );
			
			tmpObstacle.r = r;
			tmpObstacle.t = t;
			tmpObstacle.x = theGoalPercept.posts[i].positionOnField.x;
			tmpObstacle.y = theGoalPercept.posts[i].positionOnField.y;
			tmpObstacle.behind = false;
			tmpObstacle.costs = 500;
			tmpObstacle.usgoal = true;
			
			obstacles.push_back( tmpObstacle );
		}
	}
	
	// robot positions
	for ( RobotsModel::RCIt it = theRobotsModel.robots.begin(); it != theRobotsModel.robots.end(); ++it)
	{
		int r, t;
		getPolarPosition( r, t, it->relPosOnField.x, it->relPosOnField.y );
		
		tmpObstacle.r = r;
		tmpObstacle.t = t;
		tmpObstacle.x = it->relPosOnField.x;
		tmpObstacle.y = it->relPosOnField.y;
		tmpObstacle.behind = false;
		tmpObstacle.costs = 800;
		tmpObstacle.usgoal = false;
		
		obstacles.push_back( tmpObstacle );
	}
	
	// ultrasonic sensors
	int distance = 1000; // threshold
	
	// left sensor
	if ( theObstacleModel.distanceToLeftObstacle < distance )
		InsertUsMeasurement(theObstacleModel.distanceToLeftObstacle, -5.f * M_PI / 180.f, 55.f * M_PI / 180.f);
	
	// right sensor
	if ( theObstacleModel.distanceToRightObstacle < distance )
		InsertUsMeasurement(theObstacleModel.distanceToRightObstacle, -55.f * M_PI / 180.f, 5.f * M_PI / 180.f);
}

void PolargridPathPlanner::InsertUsMeasurement(float distance, float fromAngle, float toAngle)
{
	obstacle tmpObstacle;
	float radius = distance + 64.f; // floor(sqrt( 40² + 50² )) mm
	int r = log( ( radius / OFFSET ) + 1 ) / LOGBASE;
	float x, y;
		
	for ( int t = getT( fromAngle ); t <= ( getT( toAngle ) + GRID_THETA ); t++ )
	{
		getCartPosition( x, y, r, ( t % GRID_THETA ) );
		
		tmpObstacle.r = r;
		tmpObstacle.t = ( t % GRID_THETA );
		tmpObstacle.x = x;
		tmpObstacle.y = y;
		tmpObstacle.behind = false;
		tmpObstacle.costs = 800;
		tmpObstacle.usgoal = true;
		
		obstacles.push_back( tmpObstacle );
	}
}

void PolargridPathPlanner::InitRobotObstacle() {
	obstacle tmpObstacle;
	const int size = 2;
	
	// raise costs with regard to the robots movement direction and speed
	float speedX, speedY, faktor;
	int t, invT;
	speedX = theMotionInfo.walkRequest.speed.translation.x;
	speedY = theMotionInfo.walkRequest.speed.translation.y;
	int costs = 100;
		
	if ( speedX != 0 || speedY != 0 ) {
		t = getT( atan2( speedY, speedX ) );
		invT = ( t + ( GRID_THETA / 2 ) ) % GRID_THETA;
		
		if ( ( speedX != 0 && speedY == 0 ) || ( speedX == 0 && speedY != 0 ))
			faktor = abs( speedX ) + abs( speedY );
		else
			faktor = sqrt( speedX * speedX + speedY * speedY );
		
		for ( int k = ( invT - size); k <= ( invT + size); k++ ) {
			for ( int l = 1; l <= size; l++ ) {
				int m;
				if ( k < 0 ) {
					m = GRID_THETA + k;
				} 
				else {
					if ( k >= GRID_THETA ) {
						m = k - GRID_THETA;
					}
					else {
						m = k;
					}
				}
				
				tmpObstacle.r = l;
				tmpObstacle.t = m;
				tmpObstacle.behind = true;
				tmpObstacle.costs = ( costs * faktor );
				tmpObstacle.usgoal = false;
			
				obstacles.push_back( tmpObstacle );				
			}
		}
	}
}

void PolargridPathPlanner::InsertObstacles() {
	obstacle tmpObstacle;

	// expand obstacles
	int size, samecostssize, completesize;
	int length = obstacles.size();
	float decrease;
	int decrfactor;
	const int robotsize = int( 0.5f + ROBOTRADIUS / MIN_CELL_SIZE );
	
	for ( int i = 0; i < length; i++ ) {
		if ( obstacles[i].behind == false ) {
			
			// different radii
			if ( obstacles[i].usgoal == true ) {
				size = int( 0.5f + USGOALRADIUS / MIN_CELL_SIZE );
				samecostssize = size + robotsize;
				completesize = samecostssize + robotsize;
			}
			else {
				size = int( 0.5f + ROBOTRADIUS / MIN_CELL_SIZE );
				samecostssize = 2 * size;
				completesize = 3 * size;
			}
			
			for ( int j = -completesize; j <= completesize; j++ ) {
				for ( int k = -completesize; k <= completesize; k++ ) {
					float nx, ny;
					nx = obstacles[i].x + j * MIN_CELL_SIZE;
					ny = obstacles[i].y + k * MIN_CELL_SIZE;
					
					if ( j == 0 && k == 0 )
						continue;
					
					int r, t;
					getPolarPosition( r, t, nx, ny );
					
					if ( r < 0 || r >= GRID_RHO || t < 0 || t >= GRID_THETA )
						continue;
					
					tmpObstacle.r = r;
					tmpObstacle.t = t;
					tmpObstacle.x = nx;
					tmpObstacle.y = ny;
					tmpObstacle.behind = true;
					
					if ( j < -samecostssize || j > samecostssize || k < -samecostssize || k > samecostssize ) {
						// linear decreasing costs
						if ( abs(j) > abs(k))
							decrfactor = abs(j) - samecostssize;
						else
							decrfactor = abs(k) - samecostssize;
						decrease = obstacles[i].costs / robotsize;
						tmpObstacle.costs = obstacles[i].costs - decrfactor * decrease;
					} 
					else {
						// same costs
						tmpObstacle.costs = obstacles[i].costs;
					}
				
					obstacles.push_back( tmpObstacle );
				}
			}
		}
	}
	
	// insert obstacles into grid
	for ( unsigned int i = 0; i < obstacles.size(); i++ ) {
		
		if ( obstacles[i].r < 0 || obstacles[i].r >= GRID_RHO || obstacles[i].t < 0 || obstacles[i].t >= GRID_THETA )
			continue;
		
		(*grid)[obstacles[i].r][obstacles[i].t] += obstacles[i].costs / ( (obstacles[i].r+1));
	}

	COMPLEX_DRAWING("module:PolargridPathPlanner:obstacles", {
		for(size_t i = 0; i < obstacles.size(); i++)
		{
			if(obstacles[i].r < 0 || obstacles[i].r >= GRID_RHO || obstacles[i].t < 0 || obstacles[i].t >= GRID_THETA)
				continue;
			int color = ( obstacles[i].costs / ( (obstacles[i].r+1) * (obstacles[i].r+1) ) * 2 );
			if ( color > 235 )
				color = 235;

			float x, y;
			getCartPosition( x, y, obstacles[i].r, obstacles[i].t );

			LARGE_DOT("module:PolargridPathPlanner:obstacles", x, y, ColorRGBA( color, color, color ), ColorRGBA( color, color, color ));
		}
	});
}

float PolargridPathPlanner::lookupCosts( int r, int t, int nr, int nt ) const
{
	int fromR = std::min(r, nr);
	int rdiff = abs(r-nr);
	int tdiff = abs(t-nt);
	if(tdiff == GRID_THETA - 1)
		tdiff = 1;
	return lookupcosts[fromR][rdiff+2*tdiff-1];
}

float PolargridPathPlanner::getCosts( int r, int t, int nr, int nt ) const
{
	float fromX, fromY, toX, toY;
	getCartPosition(fromX, fromY, r, t);
	getCartPosition(toX, toY, nr, nt);
	
	float distX = toX - fromX;
	float distY = toY - fromY;
	
	return sqrtf( distX * distX + distY * distY );
}

void PolargridPathPlanner::PlanPath( PathPlannerData& pathPlannerData, float goal_r, float goal_t ) {
	std::priority_queue< gridPosition > priorityqueue;
	std::vector< gridPosition > visited;
	
	gridPosition currPos;
	gridPosition tmpPos;
	gridPosition pathPos;
	
	currPos.grid = tmpPos.grid = pathPos.grid = &score;
	
	float costs;
	float heuristic;
	float scores;
	bool reachedGoal = false;
	bool reachedStart = false;
	
	// Startnode (0,0)
	tmpPos.r = tmpPos.t = tmpPos.scores = tmpPos.costs = 0;
	priorityqueue.push( tmpPos );
	
	while ( !priorityqueue.empty() ) {
		currPos = priorityqueue.top();
		priorityqueue.pop();
		visited.push_back( currPos );
		
		if ( currPos.r == goal_r && currPos.t == goal_t ) {
			reachedGoal = true;
			break;
		}
		
		// expand node
		for ( int i = ( currPos.r - 1 ); i <= ( currPos.r + 1 ); i++ ) {
			for ( int j = ( currPos.t - 1 ); j <= ( currPos.t + 1 ); j++ ) {
				int k;
				
				// theta is continuous
				if ( j == -1 ) {
					k = GRID_THETA - 1;
				} 
				else {
					if ( j == GRID_THETA ) {
						k = 0;
					}
					else {
						k = j;
					}
				}
				
				if ( i < 0 || i >= GRID_RHO || k < 0 || k >= GRID_THETA || ( i == currPos.r && k == currPos.t ) ) {
					continue;
				}
				
				if ( (*closed)[i][k] == 1 )
					continue;
				
				tmpPos.r = i;
				tmpPos.t = k;
				
				// calculate costs
				costs = currPos.costs + lookupCosts( currPos.r, currPos.t, tmpPos.r, tmpPos.t ) + (*grid)[i][k];
				tmpPos.costs = costs;
					
				// calculate heuristic
				float goalX, goalY, tmpPosX, tmpPosY;
				getCartPosition( goalX, goalY, goal_r, goal_t );
				getCartPosition( tmpPosX, tmpPosY, tmpPos.r, tmpPos.t );
				heuristic = sqrtf( ( goalX - tmpPosX ) * ( goalX - tmpPosX ) + ( goalY - tmpPosY ) * ( goalY - tmpPosY ) );
				
				scores = costs + heuristic;
				
				// visualize expansions of A*
				/*
				float currPosX, currPosY;
				getCartPosition( currPosX, currPosY, currPos.r, currPos.t );
				LINE("module:PolargridPathPlanner:expansion", currPosX, currPosY, tmpPosX, tmpPosY, 10, Drawings::ps_solid, ColorClasses::red);
				*/
				
				if ( ( (*score)[i][k] != 0 ) && ( (*score)[i][k] <= scores ) )
					continue;
				
				// parent node
				tmpPos.pr = currPos.r;
				tmpPos.pt = currPos.t;
				
				(*score)[i][k] = scores;
				tmpPos.scores = scores;
				priorityqueue.push( tmpPos );

				// visualize expansions of A* after testing scores
				/*
				float currPosX, currPosY;
				getCartPosition( currPosX, currPosY, currPos.r, currPos.t );
				LINE("module:PolargridPathPlanner:expansion", currPosX, currPosY, tmpPosX, tmpPosY, 10, Drawings::ps_solid, ColorClasses::red);
				*/
			}
		}
		
		(*closed)[currPos.r][currPos.t] = 1;
	}
	
	// backtrack path
	if ( reachedGoal && !visited.empty() ) {	
		pathPos = visited.back();
		visited.pop_back();
		pathPlannerData.path.clear();
		float waypoint_x, waypoint_y;
		getCartPosition( waypoint_x, waypoint_y, pathPos.r, pathPos.t );
		pathPlannerData.path.push_back( Vector2<float>(waypoint_x, waypoint_y) );
		
		while ( !reachedStart ) {			
			float scores = FLT_MAX;
			for ( unsigned int i = 0; i < visited.size(); i++ ) {		
				if ( visited[i].r == pathPos.pr && visited[i].t == pathPos.pt ) {
					if ( visited[i].scores < scores ) {
						scores = visited[i].scores;
						pathPos = visited[i];
					}
				}
			}
			
			getCartPosition( waypoint_x, waypoint_y, pathPos.r, pathPos.t );
			pathPlannerData.path.push_back( Vector2<float>(waypoint_x, waypoint_y) );
			if ( pathPos.r == 0 ) {
				reachedStart = true;
				break;
			}
		}
	}
}
 
MAKE_MODULE(PolargridPathPlanner, Modeling)
