#ifndef __PolargridPathPlanner_h__
#define __PolargridPathPlanner_h__

#include "Tools/Module/Module.h"

#include "Representations/Modeling/ObstacleModel.h"
#include "Representations/MotionControl/OdometryData.h"
#include "Representations/Perception/GoalPercept.h"
#include "Representations/Modeling/RobotsModel.h"
#include "Representations/Infrastructure/FrameInfo.h"
#include "Representations/MotionControl/MotionInfo.h"
#include "Representations/Modeling/RobotsModel.h"
#include "Representations/Modeling/PathPlannerData.h"
#include "Representations/Modeling/PathPlannerControl.h"

#include <math.h>
#include <float.h>
#include <limits.h>

#define _USE_MATH_DEFINES

#define RADIUS 7211.f // in mm
#define GRID_RHO 16 // number of circles
#define GRID_THETA 16 // number of fragmentations -> best is a multiple of 4
#define MIN_CELL_SIZE 100.f // size of smallest radius in mm

#define BASE 1.1789
const float LOGBASE = log( BASE );
const float OFFSET = MIN_CELL_SIZE / ( BASE - 1 );

#define ROBOTRADIUS 150.f
#define USGOALRADIUS 55.f

MODULE(PolargridPathPlanner)
	REQUIRES(ObstacleModel)
	REQUIRES(OdometryData)
	REQUIRES(GoalPercept)
	REQUIRES(FrameInfo)
	REQUIRES(MotionInfo)
	REQUIRES(RobotsModel)
	USES(PathPlannerControl)
	PROVIDES_WITH_MODIFY(PathPlannerData)
END_MODULE


class PolargridPathPlanner : public PolargridPathPlannerBase {
private:
	void update( PathPlannerData& pathPlannerData );

public:
	PolargridPathPlanner();
	typedef float GRID_TYPE;
	typedef GRID_TYPE (*GRID)[GRID_RHO][GRID_THETA];
	GRID grid; // obstacles
	GRID score;	// costs + heuristics
	GRID closed; // visited cells
	
	// lookup tables
	float lookuprho[GRID_RHO+1];
	float lookuprhohalf[GRID_RHO+1];
	float lookuptheta[GRID_THETA];
	float lookupsintheta[GRID_THETA];
	float lookupcostheta[GRID_THETA];
	float lookupcosts[GRID_RHO+1][3];
 	
	/** determines the target position in log-polar coordinates and initialises it in the grid */ 
	void InitTarget( int& r, int& t, float target_x, float target_y );
	/** pushes the perceived goal posts, the positions of the robots and the measurements of both ultrasonic sensors into the obstacle vector */
	void InitObstacles();
	/** initialize artificial obstacle behind the robot */
	void InitRobotObstacle();
	/** expands obstacles, insert them into the grid and draw them into the world state */
	void InsertObstacles();
	void InsertUsMeasurement(float distance, float fromAngle, float toAngle);
	/** initialize each grid with zeros */
	void InitGrid();
	/** plan path to target with A* algorithm, backtracking */
	void PlanPath( PathPlannerData& pathPlannerData, float goal_r, float goal_t );
	/** Draws the found path into the world state */
	void DrawPath(const PathPlannerData& pathPlannerData);
	/** Draws the log-polar grid structure into the world state */
	void DrawGrid();
	/** initialize lookup tables */
	void InitLookup();
	/** calculate step costs (Euclidean distance between cell centers) */
	float getCosts( int r, int t, int nr, int nt ) const;
	/** lookup precalculated step costs */
	float lookupCosts( int r, int t, int nr, int nt ) const;
	
	void updatePolargridPathPlanner( float target_x, float target_y, PathPlannerData& pathPlannerData );
	
	/** conversion from cartesian coordinates to log-polar coordinates */
	void getPolarPosition ( int& r, int& t, float x, float y );

	/** conversion from log-polar coordinates to cartesian coordinates */
	inline void getCartPosition ( float& x, float& y, int r, int t ) const
	{
		x = lookuprhohalf[r] * lookupcostheta[t];
		y = lookuprhohalf[r] * lookupsintheta[t];
	}
	
	/** calculate thete from discretization t */
	inline float getTheta ( int t ) {
		return t / ( GRID_THETA / ( 2 * M_PI ) );
	}
	
	/** discretize rho */
	inline int getR ( float rho ) { return rho; }
	
	/** discretize theta */	
	int getT ( float theta );
	
	struct gridPosition {
		int r, t; // grid coordinates
		int pr, pt; // parent coordinates
		GRID* grid;
		float scores;
		float costs;
		
		// operator for priority queue order
		bool operator<( const gridPosition &a ) const {
			return scores > a.scores;
		}
	};
	
	struct obstacle {
		int r, t; // grid coordinates
		float x, y; // cartesian coordinates
		float costs; // costs of the obstacle
		bool behind; // is artificial obstacle behind the robot
		bool usgoal; // distinguish between goal posts/ultrasonic sensor measurements and robots
	};
	
	std::vector< obstacle > obstacles;
};

inline int PolargridPathPlanner::getT ( float theta ) {
	int t;
	if ( theta < 0 )
		theta += 2 * M_PI;
	if ( theta != 2 * M_PI )
		t = ( GRID_THETA/( 2 * M_PI ) ) * theta + 0.5;
	else
		t = 0;
	return t;
}

inline void PolargridPathPlanner::DrawPath(const PathPlannerData& pathPlannerData) {
	COMPLEX_DRAWING("module:PolargridPathPlanner:path", {
		if(pathPlannerData.path.empty())
			return;
		float lastX = pathPlannerData.path.front().x;
		float lastY = pathPlannerData.path.front().y;
		for(std::vector< Vector2<float> >::const_iterator it = pathPlannerData.path.begin(); it != pathPlannerData.path.end(); ++it)
		{
			LARGE_DOT("module:PolargridPathPlanner:path", it->x, it->y, ColorClasses::yellow, ColorClasses::yellow);
			LINE("module:PolargridPathPlanner:path", lastX, lastY, it->x, it->y, 10, Drawings::ps_solid, ColorClasses::black);
			lastX = it->x;
			lastY = it->y;
		}
	}
	);
}

inline void PolargridPathPlanner::DrawGrid() {
	COMPLEX_DRAWING("module:PolargridPathPlanner:grid", {	
		for ( int i = 0; i <= GRID_RHO; i++ )
		{
			float radius = lookuprho[i];
			CIRCLE("module:PolargridPathPlanner:grid", 0, 0, radius, 10, Drawings::ps_solid, ColorClasses::black, Drawings::bs_null, ColorClasses::black);
		}

		for ( int j = 0; j < GRID_THETA; j++ )
		{
			float theta = ( j + 0.5 ) * ( 2 * M_PI / GRID_THETA );

			float toX = lookuprho[GRID_RHO] * cos( theta );
			float toY = lookuprho[GRID_RHO] * sin( theta );
			LINE("module:PolargridPathPlanner:grid", 0, 0, toX, toY, 10, Drawings::ps_solid, ColorClasses::black);
		}
	}
	);
}

inline void PolargridPathPlanner::getPolarPosition (int& r, int& t, float x, float y )
{
	float rho, theta;
	
	if ( x == 0 && y == 0 ) {
		rho = 0;
	}
	else {
		
		if ( x != 0 && x < MIN_CELL_SIZE && x > -MIN_CELL_SIZE ) {
			if ( x > 0 )
				x = MIN_CELL_SIZE;
			else
				x = -MIN_CELL_SIZE;
		}
		if ( y != 0 && y < MIN_CELL_SIZE && y > -MIN_CELL_SIZE ) {
			if ( y > 0 )
				y = MIN_CELL_SIZE;
			else
				y = -MIN_CELL_SIZE;
		}
		
		rho = log( ( sqrt( x * x + y * y ) / OFFSET ) + 1 ) / LOGBASE;
	}
	theta = atan2( y, x );
	
	r = getR( rho );
	t = getT( theta );
}

#endif
